#include "pack.h"
#include "util.h"
#include <immintrin.h>
#include <stdio.h>
#include <stdlib.h>

// pack and unpack mat R, in which each element has two bits
void pack_mat_r(uint8_t *buf, const int16_t *data)
{
    for (size_t i = 0; i < PARAM_R_BYTES; i++)
    {
        buf[i] = ((data[0] & 0x03) << 6) | ((data[1] & 0x03) << 4) | ((data[2] & 0x03) << 2) | (data[3] & 0x03);
        data += 4;
    }
}

void unpack_mat_r_avx2(int16_t *data, const uint8_t *buf, int buf_len)
{
    for (int i = 0; i < buf_len / 32; i++)
    {
        __m256i a0, a1, a2, a3, a4, a5, a6, a7, b0, b1, b2, b3, b4, b5, b6, b7;
        a0 = _mm256_loadu_si256((__m256i *)(buf));
        a1 = _mm256_slli_epi16(a0, 2);
        a2 = _mm256_slli_epi16(a0, 4);
        a3 = _mm256_slli_epi16(a0, 6);
        a4 = _mm256_slli_epi16(a0, 8);
        a5 = _mm256_slli_epi16(a0, 10);
        a6 = _mm256_slli_epi16(a0, 12);
        a7 = _mm256_slli_epi16(a0, 14);

        a0 = _mm256_srai_epi16(a0, 14);
        a1 = _mm256_srai_epi16(a1, 14);
        a2 = _mm256_srai_epi16(a2, 14);
        a3 = _mm256_srai_epi16(a3, 14);
        a4 = _mm256_srai_epi16(a4, 14);
        a5 = _mm256_srai_epi16(a5, 14);
        a6 = _mm256_srai_epi16(a6, 14);
        a7 = _mm256_srai_epi16(a7, 14);

        b0 = _mm256_unpacklo_epi16(a0, a1);
        b1 = _mm256_unpackhi_epi16(a0, a1);
        b2 = _mm256_unpacklo_epi16(a2, a3);
        b3 = _mm256_unpackhi_epi16(a2, a3);
        b4 = _mm256_unpacklo_epi16(a4, a5);
        b5 = _mm256_unpackhi_epi16(a4, a5);
        b6 = _mm256_unpacklo_epi16(a6, a7);
        b7 = _mm256_unpackhi_epi16(a6, a7);

        a0 = _mm256_unpacklo_epi32(b0, b2);
        a1 = _mm256_unpackhi_epi32(b0, b2);
        a2 = _mm256_unpacklo_epi32(b1, b3);
        a3 = _mm256_unpackhi_epi32(b1, b3);
        a4 = _mm256_unpacklo_epi32(b4, b6);
        a5 = _mm256_unpackhi_epi32(b4, b6);
        a6 = _mm256_unpacklo_epi32(b5, b7);
        a7 = _mm256_unpackhi_epi32(b5, b7);

        b0 = _mm256_unpacklo_epi64(a4, a0);
        b1 = _mm256_unpackhi_epi64(a4, a0);
        b2 = _mm256_unpacklo_epi64(a5, a1);
        b3 = _mm256_unpackhi_epi64(a5, a1);
        b4 = _mm256_unpacklo_epi64(a6, a2);
        b5 = _mm256_unpackhi_epi64(a6, a2);
        b6 = _mm256_unpacklo_epi64(a7, a3);
        b7 = _mm256_unpackhi_epi64(a7, a3);

        a0 = _mm256_permute2f128_si256(b0, b1, 0x20);
        a1 = _mm256_permute2f128_si256(b0, b1, 0x31);
        a2 = _mm256_permute2f128_si256(b2, b3, 0x20);
        a3 = _mm256_permute2f128_si256(b2, b3, 0x31);
        a4 = _mm256_permute2f128_si256(b4, b5, 0x20);
        a5 = _mm256_permute2f128_si256(b4, b5, 0x31);
        a6 = _mm256_permute2f128_si256(b6, b7, 0x20);
        a7 = _mm256_permute2f128_si256(b6, b7, 0x31);

        _mm256_storeu_si256((__m256i *)data, a0);
        _mm256_storeu_si256((__m256i *)(data + 16), a2);
        _mm256_storeu_si256((__m256i *)(data + 32), a4);
        _mm256_storeu_si256((__m256i *)(data + 48), a6);
        _mm256_storeu_si256((__m256i *)(data + 64), a1);
        _mm256_storeu_si256((__m256i *)(data + 80), a3);
        _mm256_storeu_si256((__m256i *)(data + 96), a5);
        _mm256_storeu_si256((__m256i *)(data + 112), a7);
        buf = buf + 32;
        data = data + 128;
    }
}
