#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "api.h"
#include "cpucycles.h"
#include "params.h"
#include "sha3/shake.h"
#include "util.h"
#include <time.h>

#define NTESTS 10

struct Result
{
    clock_t tgen, tsign, tvrfy;
};

uint64_t TGEN0[NTESTS], TSIGN0[NTESTS], TVRFY0[NTESTS];
uint64_t TGEN1[NTESTS], TSIGN1[NTESTS], TVRFY1[NTESTS];

uint8_t pk[CRYPTO_PUBLICKEYBYTES];
uint8_t sk[CRYPTO_SECRETKEYBYTES];

int test_signature(struct Result *result, int cnt)
{
    int success;
    clock_t tgen0, tsign0, tvrfy0;
    clock_t tgen1, tsign1, tvrfy1;

    tgen0 = clock();
    TGEN0[cnt] = cpucycles();
    success = crypto_sign_keypair(pk, sk);
    TGEN1[cnt] = cpucycles();
    tgen1 = clock();

    if (success != 0)
    {
        return -1;
    }

    unsigned long long smlen, mlen1;
    const unsigned long long mlen = 32;
    uint8_t m[mlen];
    for (size_t i = 0; i < mlen; i++)
        m[i] = i;
    uint8_t sm[mlen + CRYPTO_BYTES];

    tsign0 = clock();
    TSIGN0[cnt] = cpucycles();
    success = crypto_sign(sm, &smlen, m, mlen, sk);
    TSIGN1[cnt] = cpucycles();
    tsign1 = clock();

    if (success != 0)
    {
        return -1;
    }

    if (smlen > mlen + CRYPTO_BYTES)
    {
        return -1;
    }

    tvrfy0 = clock();
    TVRFY0[cnt] = cpucycles();
    success = crypto_sign_open(m, &mlen1, sm, smlen, pk);
    TVRFY1[cnt] = cpucycles();
    tvrfy1 = clock();

    if (success != 0)
    {
        return -1;
    }

    if (mlen1 != mlen)
    {
        return -1;
    }

    for (size_t i = 0; i < mlen; i++)
        if (m[i] != i)
            return -1;

    result->tgen += tgen1 - tgen0;
    result->tsign += tsign1 - tsign0;
    result->tvrfy += tvrfy1 - tvrfy0;
    return 0;
}
struct Result ret;

int main()
{
    for (int i = 0; i < NTESTS; i++)
    {
        if (test_signature(&ret, i) != 0)
        {
            printf("Failed!");
            return 0;
        }
    }

    print_results("KeyGen:", TGEN0, TGEN1, NTESTS, ret.tgen);
    print_results("Sign:", TSIGN0, TSIGN1, NTESTS, ret.tsign);
    print_results("Verify:", TVRFY0, TVRFY1, NTESTS, ret.tvrfy);
    return 0;
}
