#include "cpucycles.h"
#include <time.h>

uint64_t cpucycles_overhead(void)
{
    uint64_t t0, t1, overhead = -1LL;
    unsigned int i;

    for (i = 0; i < 100000; i++)
    {
        t0 = cpucycles();
        __asm__ volatile("");
        t1 = cpucycles();
        if (t1 - t0 < overhead)
            overhead = t1 - t0;
    }

    return overhead;
}

//------------------------------------------------------------------
static int cmp_uint64(const void *a, const void *b)
{
    if (*(uint64_t *)a < *(uint64_t *)b)
        return -1;
    if (*(uint64_t *)a > *(uint64_t *)b)
        return 1;
    return 0;
}

static uint64_t median(uint64_t *l, size_t llen)
{
    qsort(l, llen, sizeof(uint64_t), cmp_uint64);

    if (llen % 2)
        return l[llen / 2];
    else
        return (l[llen / 2 - 1] + l[llen / 2]) / 2;
}

static uint64_t average(uint64_t *t, size_t tlen)
{
    size_t i;
    uint64_t acc = 0;

    for (i = 0; i < tlen; i++)
        acc += t[i];

    return acc / tlen;
}

void print_singleres(const char *s, uint64_t t0, uint64_t t1)
{
    size_t i;
    static uint64_t overhead = -1;

    if (overhead == (uint64_t)-1)
        overhead = cpucycles_overhead();

    uint64_t t = 0;
    t = t1 - t0 - overhead;
    printf("%s %lucycles/ticks\n", s, t);
}

void print_results(const char *s, uint64_t *t0, uint64_t *t1, size_t tlen, clock_t clk)
{
    size_t i;
    static uint64_t overhead = -1;

    if (tlen < 2)
    {
        fprintf(stderr, "ERROR: Need a least two cycle counts!\n");
        return;
    }

    if (overhead == (uint64_t)-1)
        overhead = cpucycles_overhead();

    // tlen--;
    uint64_t t[tlen];
    for (i = 0; i < tlen; ++i)
        t[i] = t1[i] - t0[i] - overhead;

    printf("%10s\t", s);
    printf("med (cycles/ticks):\t%12llu\t", (unsigned long long)median(t, tlen));
    printf("avg (cycles/ticks):\t%12llu\t", (unsigned long long)average(t, tlen));
    printf("time (s):\t%8.6lf\n", 1.0 * clk / (tlen * CLOCKS_PER_SEC));
}
